#include "main-view-controller.h"

#include <QTimer>

#include "spice-remote-viewer.h"

namespace {

const QString kServerURL = "serverURL";
const QString kVirtualMachineID = "vmid";
const QString kUsername = "username";
const QString kPassword = "password";
const QString kRememberMe = "rememberMe";
const QString kIgnoreSslErrors = "ignoreSslErrors";

const QString kServerURLIsEmpty = "Server URL field is empty";
const QString kUsernameIsEmpty = "User name field is empty";
const QString kPasswordIsEmpty = "Password field is empty";

const QString kRemoteViewerLaunchFailed = "Could not launch remote viewer";

const int kVirtualMachineStatusCheckTimeout = 1000;

} // namespace

namespace Basealt {

MainViewController::MainViewController( std::unique_ptr<ISettingsManager> manager )
    : _settingsManager( std::move( manager ) ),
      _view(),
      _client(),
      _virtualMachine(),
      _serverURL(),
      _virtualMachineID( 0 ),
      _username(),
      _password(),
      _rememberMe( false ),
      _ignoreSslErrors( false )
{
}

void MainViewController::bind( const std::shared_ptr<MainView>& view )
{
    _view = view;

    if( !view )
    {
        return;
    }

    auto settingsView = view->settingsView();

    connect( settingsView,
             &SettingsView::serverURLChanged,
             this,
             &MainViewController::onServerURLChanged );

    connect( settingsView,
             &SettingsView::virtualMachineIDChanged,
             this,
             &MainViewController::onVirtualMachineIDChanged );

    connect( settingsView,
             &SettingsView::usernameChanged,
             this,
             &MainViewController::onUsernameChanged );

    connect( settingsView,
             &SettingsView::passwordChanged,
             this,
             &MainViewController::onPasswordChanged );

    connect( settingsView,
             &SettingsView::rememberMeStateChanged,
             this,
             &MainViewController::onRememberMeStateChanged );

    connect( settingsView,
             &SettingsView::ignoreSslErrorsStateChanged,
             this,
             &MainViewController::onIgnoreSslErrorsStateChanged );

    connect( settingsView,
             &SettingsView::connectButtonClicked,
             this,
             &MainViewController::onConnectButtonClicked );

    auto processView = view->processView();

    connect( processView,
             &ProcessView::cancelButtonClicked,
             this,
             &MainViewController::onCancelButtonClicked );

    _rememberMe = _settingsManager->get( kRememberMe ).toBool();

    if( _rememberMe )
    {
        restoreViewState();
    }
}

void MainViewController::bind( ProxmoxClient* client )
{
    connect( client,
             &ProxmoxClient::loginFailed,
             this,
             &MainViewController::onErrorOccured );

    connect( client,
             &ProxmoxClient::loginSucceeded,
             this,
             &MainViewController::onLoginSucceeded );

    connect( client,
             &ProxmoxClient::nodesFetched,
             this,
             &MainViewController::onNodesFetched );

    connect( client,
             &ProxmoxClient::nodesFetchFailed,
             this,
             &MainViewController::onErrorOccured );
}

void MainViewController::bind( VirtualMachine* virtualMachine )
{
    connect( virtualMachine,
             &VirtualMachine::statusFetched,
             this,
             &MainViewController::onVirtualMachineStateFetched );

    connect( virtualMachine,
             &VirtualMachine::statusFetchFailed,
             this,
             &MainViewController::onErrorOccured );

    connect( virtualMachine,
             &VirtualMachine::spiceProxyFetched,
             this,
             &MainViewController::onVirtualMachineSpiceProxyFetched );

    connect( virtualMachine,
             &VirtualMachine::spiceProxyFetchFailed,
             this,
             &MainViewController::onErrorOccured );

    connect( virtualMachine,
             &VirtualMachine::statusSet,
             this,
             &MainViewController::onVirtualMachineStatusSet );

    connect( virtualMachine,
             &VirtualMachine::setStatusFailed,
             this,
             &MainViewController::onErrorOccured );
}

void MainViewController::onServerURLChanged( const QString& serverURL )
{
    _serverURL = serverURL;
    
    if( _rememberMe )
    {
        _settingsManager->set( kServerURL, _serverURL );
    }
}

void MainViewController::onVirtualMachineIDChanged( const QString& vmid )
{
    _virtualMachineID = static_cast<size_t>( vmid.toUInt() );

    if( _rememberMe )
    {
        _settingsManager->set( kVirtualMachineID, vmid );
    }
}

void MainViewController::onUsernameChanged( const QString& username )
{
    _username = username;

    if( _rememberMe )
    {
        _settingsManager->set( kUsername, _username );
    }
}

void MainViewController::onPasswordChanged( const QString& password )
{
    _password = password;
}

void MainViewController::onRememberMeStateChanged( bool isChecked )
{
    _rememberMe = isChecked;

    _settingsManager->set( kRememberMe, _rememberMe );
}

void MainViewController::onIgnoreSslErrorsStateChanged( bool isChecked )
{
    _ignoreSslErrors = isChecked;

    if( _rememberMe )
    {
        _settingsManager->set( kIgnoreSslErrors, _ignoreSslErrors );
    }
}

void MainViewController::onLoginSucceeded()
{
    if( const auto view = _view )
    {
        view->processView()->setState( ApplicationState::FetchingNodes );
    }

    if( const auto client = _client )
    {
        client->fetchNodes();
    }
}

void MainViewController::onNodesFetched( const std::vector<std::shared_ptr<Node>>& nodes )
{
    if( const auto view = _view )
    {
        view->processView()->setState( ApplicationState::FetchingVMs );
    }

    for( const auto& node : nodes )
    {
        connect( node.get(),
                 &Node::virtualMachinesFetched,
                 this,
                 &MainViewController::onVirtualMachinesFetched );

        connect( node.get(),
                 &Node::virtualMachinesFetchFailed,
                 this,
                 &MainViewController::onErrorOccured );

        node->fetchVirtualMachines();
    }
}

void MainViewController::onVirtualMachinesFetched( const std::vector<std::shared_ptr<VirtualMachine>>& vms )
{
    const auto vmid = _virtualMachineID;

    auto rule = [vmid]( const std::shared_ptr<VirtualMachine>& vm )
    {
        return vm->id() == vmid;
    };

    const auto found = std::find_if( vms.begin(), vms.end(), rule );

    if( found != vms.end() )
    {
        _virtualMachine = *found;
        bind( _virtualMachine.get() );

        if( const auto view = _view )
        {
            view->processView()->setState( ApplicationState::FetchingVMState );
        }

        if( const auto virtualMachine = _virtualMachine )
        {
            virtualMachine->fetchStatus();
        }
    }
}

void MainViewController::onVirtualMachineStateFetched( VirtualMachine::Status status )
{
    if( status == VirtualMachine::Status::Running )
    {
        if( const auto view = _view )
        {
            view->processView()->setState( ApplicationState::FetchingSpiceTicket );
        }

        if( const auto virtualMachine = _virtualMachine )
        {
            virtualMachine->fetchSpiceProxy();
        }
    }
    else
    {
        if( const auto view = _view )
        {
            view->processView()->setState( ApplicationState::StartingVM );
        }

        if( const auto virtualMachine = _virtualMachine )
        {
            virtualMachine->setStatus( VirtualMachine::Status::Running );
        }
    }
}

void MainViewController::onVirtualMachineStatusSet()
{
     QTimer::singleShot( kVirtualMachineStatusCheckTimeout,
                         this,
                         &MainViewController::checkVirtualMachineStatus );
}

void MainViewController::checkVirtualMachineStatus()
{
    if( const auto virtualMachine = _virtualMachine )
    {
        virtualMachine->fetchStatus();
    }
}

void MainViewController::onVirtualMachineSpiceProxyFetched( const QString& spiceProxyFilePath )
{
    const auto view = _view;

    if( view )
    {
        view->processView()->setState( ApplicationState::RunningViewer );
    }

    SpiceRemoteViewer viewer( spiceProxyFilePath );

    const auto opened = viewer.open();

    if( view )
    {
        view->hideSpinner();
    }

    if( opened )
    {
        emit done();
    }
    else
    {
        if( view )
        {
            view->reportError( kRemoteViewerLaunchFailed );
        }
    }
}

void MainViewController::onErrorOccured( const QString& reason )
{
    if( const auto view = _view )
    {
        view->hideSpinner();
        view->reportError( reason );
    }
}

void MainViewController::onConnectButtonClicked()
{
    const auto view = _view;

    if( _serverURL.isEmpty() )
    {
        if( view )
        {
            view->reportError( kServerURLIsEmpty );
        }

        return;
    }

    _client = std::make_shared<ProxmoxClient>( _serverURL, _ignoreSslErrors );

    const auto client = _client;

    if( !client )
    {
        return;
    }

    bind( client.get() );

    if( _username.isEmpty() )
    {
        if( view )
        {
            view->reportError( kUsernameIsEmpty );
        }

        return;
    }

    if( _password.isEmpty() )
    {
        if( view )
        {
            view->reportError( kPasswordIsEmpty );
        }

        return;
    }

    if( view )
    {
        view->showSpinner();
        view->processView()->setState( ApplicationState::Logging );
    }

    client->login( _username, _password );
}

void MainViewController::onCancelButtonClicked()
{
    _client = nullptr;
    _virtualMachine = nullptr;

    if( const auto view = _view )
    {
        view->hideSpinner();
    }
}

void MainViewController::restoreViewState()
{
    _serverURL = _settingsManager->get( kServerURL ).toString();

    const auto vmid = _settingsManager->get( kVirtualMachineID ).toUInt();

    _virtualMachineID = static_cast<size_t>( vmid );
    _username = _settingsManager->get( kUsername ).toString();
    _ignoreSslErrors = _settingsManager->get( kIgnoreSslErrors ).toBool();

    const auto view = _view;
    
    if( view )
    {
        auto settingsView = view->settingsView();

        settingsView->setServerURL( _serverURL );
        settingsView->setVirtualMachineID( _virtualMachineID );
        settingsView->setUsername( _username );
        settingsView->setRememberMeState( true );
        settingsView->setIgnoreSslErrorsState( _ignoreSslErrors );
    }
}

} // namespace Basealt
