#include "proxmox-client.h"

#include <QUrlQuery>

#include "proxmox-json-parser.h"

namespace {

const char kSeparator = '/';

const QString kTicketEndpoint = "/access/ticket";
const QString kUsername = "username";
const QString kPassword = "password";

const QString kNodesEndpoint = "/nodes";

const QString kLoginNetworkErrorMessage = "Login request failed: ";
const QString kLoginJsonParsingErrorMessage = "Could not parse login response json";
const QString kLoginJsonBadFormatMessage = "Login response json bad format: ";
const QString kLoginInvalidURLMessage = "Login URL is invalid: ";

const QString kNodesJsonBadFormatMessage = "Nodes response json bad format: ";

std::unique_ptr<Basealt::User> processLoginReply( QNetworkReply* reply )
{
    try
    {
        const auto tokenResponseData = Basealt::ProxmoxJsonParser::parseTokenResponse( reply );

        return std::make_unique<Basealt::User>( tokenResponseData.username,
                                                tokenResponseData.ticket,
                                                tokenResponseData.token );
    }
    catch( const Basealt::TokenParsingException& ex )
    {
        throw Basealt::ProxmoxClientException( kLoginJsonBadFormatMessage
                                                + ex.what() );
    }
}

} // namespace

namespace Basealt {

ProxmoxClient::ProxmoxClient( const QString& serverURL, bool ignoreSslErrors )
    : _restClient( std::make_shared<ProxmoxRestClient>( serverURL,
                                                        ignoreSslErrors ) ),
      _loginReply( nullptr ),
      _fetchNodesReply( nullptr ),
      _user(),
      _nodes()
{
    connect( _restClient.get(),
             &ProxmoxRestClient::finished,
             this,
             &ProxmoxClient::onReplyFinished );
}

void ProxmoxClient::login( const QString& username, 
                           const QString& password )
{
    if( _loginReply )
    {
        return;
    }

    QUrlQuery postData;
    postData.addQueryItem( kUsername, QUrl::toPercentEncoding( username ) );
    postData.addQueryItem( kPassword, QUrl::toPercentEncoding( password ) );

    const auto data = postData.query().toUtf8();

    _loginReply = _restClient->post( kTicketEndpoint, data );
}

void ProxmoxClient::fetchNodes()
{
    if( !_user || _fetchNodesReply )
    {
        return;
    }

    _fetchNodesReply = _restClient->get( kNodesEndpoint );
}

void ProxmoxClient::onReplyFinished( QNetworkReply* reply )
{
    if( _loginReply && _loginReply == reply  )
    {
        try
        {
            _user = processLoginReply( reply );
            _restClient->setUser( _user );

            emit loginSucceeded();
        }
        catch( const std::exception& ex )
        {
            emit loginFailed( ex.what() );
        }

        _loginReply = nullptr;
        reply->deleteLater();

        return;
    }

    if( _fetchNodesReply && _fetchNodesReply == reply )
    {
        try
        {
            _nodes = processFetchNodesReply( reply );
            emit nodesFetched( _nodes );
        }
        catch( const std::exception& ex )
        {
            emit nodesFetchFailed( ex.what() );
        }

        _fetchNodesReply = nullptr;
        reply->deleteLater();

        return;
    }
}

std::vector<std::shared_ptr<Node>> ProxmoxClient::processFetchNodesReply( QNetworkReply* reply )
{
    try
    {
        const auto nodeIDs = ProxmoxJsonParser::parseNodesResponse( reply );

        std::vector<std::shared_ptr<Node>> result;
        result.reserve( nodeIDs.size() );

        for( const auto id : nodeIDs )
        {
            result.emplace_back( std::make_shared<Node>( _restClient, id ) );
        }

        return result;
    }
    catch( const NodesParsingException& ex )
    {
        throw ProxmoxClientException( kNodesJsonBadFormatMessage
                                      + ex.what() );
    }
}

} // namespace Basealt
