#include "proxmox-json-parser.h"

#include <unordered_map>
#include <iostream>

#include <QJsonObject>
#include <QJsonArray>
#include <QJsonDocument>
#include <QFile>
#include <QDir>

#include "errors.h"

namespace {

const QString kData = "data";
const QString kToken = "CSRFPreventionToken";
const QString kTicket = "ticket";
const QString kUsername = "username";

const QString kNode = "node";
const QString kVmID = "vmid";
const QString kStatus = "status";

const QString kIsNull = "Json is null";
const QString kIsNotAnObject = "Json is not an object";
const QString kDataNotFound = "Data entry not found";
const QString kDataIsNotAnObject = "Data entry is not an object";
const QString kDataIsNotAnArray = "Data entry is not an array";
const QString kDataIsNotAString = "Data entry is not a string";
const QString kTicketNotFound = "Ticket entry not found";
const QString kTokenNotFound = "Token entry not found";
const QString kUsernameNotFound = "Username entry not found";
const QString kVirtualMachineIsNotAnObject = "Virtual machine json is not an object";
const QString kVmIDNotFound = "Vmid entry not found";
const QString kNodeIsNotAnObject = "Node json is not an object";
const QString kNodeNotFound = "Node entry not found";
const QString kStatusNotFound = "Status entry not found";

const QString kJsonParsingErrorMessage = "Could not parse response json: ";

const QString kSpiceProxyFileNotOpened = "Could not open spice-proxy settings file";

const QString kVirtViewerGroup = "virt-viewer";

const QString kSpiceProxyFilePath = QDir::homePath()
                                    + QDir::separator()
                                    + "spiceProxySettings";

const std::unordered_map<std::string, Basealt::VirtualMachine::Status> kStatuses
{
    { "running", Basealt::VirtualMachine::Status::Running },
    { "stopped", Basealt::VirtualMachine::Status::Stopped }
};

QJsonDocument getJson( QNetworkReply* reply )
{
    const auto networkError = reply->error();

    if( networkError != QNetworkReply::NoError )
    {
        throw Basealt::NetworkException( reply->errorString() );
    }

    QJsonParseError parsingError;
    const auto json = QJsonDocument::fromJson( reply->readAll(),
                                               &parsingError );

    if( parsingError.error != QJsonParseError::NoError )
    {
       throw Basealt::JsonParsingException( parsingError.errorString() );
    }

    return json;
}

QJsonValue getDataEntry( const QJsonDocument& json )
{
    if( json.isNull() )
    {
        throw Basealt::DataEntryParsingException( kIsNull );
    }

    if( !json.isObject() )
    {
        throw Basealt::DataEntryParsingException( kIsNotAnObject );
    }

    const auto jsonObject = json.object();

    if( !jsonObject.contains( kData ) )
    {
        throw Basealt::DataEntryParsingException( kDataNotFound );
    }

    return jsonObject[kData];
}

QJsonArray getDataEntryAsArray( const QJsonDocument& json )
{
    const auto dataValue = getDataEntry( json );

    if( !dataValue.isArray() )
    {
        throw Basealt::DataEntryParsingException( kDataIsNotAnArray );
    }

    return dataValue.toArray();
}

QJsonObject getDataEntryAsObject( const QJsonDocument& json )
{
    const auto dataValue = getDataEntry( json );

    if( !dataValue.isObject() )
    {
        throw Basealt::DataEntryParsingException( kDataIsNotAnObject );
    }

    return dataValue.toObject();
}

QString makeSettingsFromJsonObject( const QJsonObject& object )
{
    QFile spiceProxyFile( kSpiceProxyFilePath );

    if( !spiceProxyFile.open( QFile::Text | QFile::WriteOnly ) )
    {
        throw Basealt::VirtualMachineParsingException( kSpiceProxyFileNotOpened );
    }

    {
        QTextStream out( &spiceProxyFile );

        out << '[' << kVirtViewerGroup << ']' << endl;

        const auto keys = object.keys();

        for( const auto& key : keys )
        {
            const auto value = object[key];

            if( value.isString() )
            {
                out << key << '=' << value.toString() << endl;
            }
            else
            {
                out << key << '=' << value.toInt() << endl;
            }
        }
    }

    return kSpiceProxyFilePath;
}

} // namespace

namespace Basealt {
namespace ProxmoxJsonParser {

TokenData parseTokenResponse( QNetworkReply* reply )
{
    const auto dataObject = getDataEntryAsObject( getJson( reply ) );

    if( !dataObject.contains( kTicket ) )
    {
        throw TokenParsingException( kTicketNotFound );
    }

    if( !dataObject.contains( kToken ) )
    {
        throw TokenParsingException( kTokenNotFound );
    }

    if( !dataObject.contains( kUsername ) )
    {
        throw TokenParsingException( kUsernameNotFound );
    }

    return { dataObject[kUsername].toString(),
             dataObject[kTicket].toString(),
             dataObject[kToken].toString() };
}

std::vector<size_t> parseVirtualMachinesResponse( QNetworkReply* reply )
{
    const auto vms = getDataEntryAsArray( getJson( reply ) );

    std::vector<size_t> result;
    result.reserve( vms.size() );

    for( const auto& vm : vms )
    {
        if( !vm.isObject() )
        {
            throw VirtualMachineParsingException( kVirtualMachineIsNotAnObject );
        }

        const auto vmObject = vm.toObject();

        if( !vmObject.contains( kVmID ) )
        {
            throw VirtualMachineParsingException( kVmIDNotFound );
        }

        result.emplace_back( static_cast<size_t>( vmObject[kVmID].toInt() ) );
    }

    return result;
}

std::vector<QString> parseNodesResponse( QNetworkReply* reply )
{
    const auto nodes = getDataEntryAsArray( getJson( reply ) );

    std::vector<QString> result;
    result.reserve( nodes.size() );

    for( const auto& node : nodes )
    {
        if( !node.isObject() )
        {
            throw NodesParsingException( kNodeIsNotAnObject );
        }

        const auto nodeObject = node.toObject();

        if( !nodeObject.contains( kNode ) )
        {
            throw NodesParsingException( kNodeNotFound );
        }

        result.emplace_back( nodeObject[kNode].toString() );
    }

    return result;
}

VirtualMachine::Status
parseVirtualMachineCurrentStatusResponse( QNetworkReply* reply )
{
    const auto dataObject = getDataEntryAsObject( getJson( reply ) );

    if( !dataObject.contains( kStatus ) )
    {
        throw VirtualMachineParsingException( kStatusNotFound );
    }

    return kStatuses.at( dataObject[kStatus].toString().toStdString() );
}

QString parseGetSpiceProxyResponse( QNetworkReply* reply )
{
    const auto dataObject = getDataEntryAsObject( getJson( reply ) );

    return makeSettingsFromJsonObject( dataObject );
}

QString parseVirtualMachineSetStatusResponse( QNetworkReply* reply )
{
    const auto dataValue = getDataEntry( getJson( reply ) );

    if( !dataValue.isString() )
    {
        throw DataEntryParsingException( kDataIsNotAString );
    }

    return dataValue.toString();
}

} // namespace ProxmoxJsonParser
} // namespace Basealt
