#include "proxmox-node.h"

#include "proxmox-json-parser.h"
#include "errors.h"

namespace {

const QString kVirtualMachinesEndpointFormat = "/nodes/%1/qemu";

} // namespace

namespace Basealt {

Node::Node( const std::shared_ptr<ProxmoxRestClient>& restClient,
            const QString& id )
    : _restClient( restClient ),
      _id( id ),
      _virtualMachines(),
      _virtualMachinesFetchReply( nullptr )
{
    connect( _restClient.get(),
             &ProxmoxRestClient::finished,
             this,
             &Node::onReplyFinished );
}

void Node::fetchVirtualMachines()
{
    if( _virtualMachinesFetchReply )
    {
        return;
    }

    const auto url = QString( kVirtualMachinesEndpointFormat ).arg( _id );

    _virtualMachinesFetchReply = _restClient->get( url );
}

void Node::onReplyFinished( QNetworkReply* reply )
{
    if( !_virtualMachinesFetchReply
            || _virtualMachinesFetchReply != reply )
    {
        return;
    }

    try
    {
        _virtualMachines = processVirtualMachinesFetchReply( reply );
        emit virtualMachinesFetched( _virtualMachines );
    }
    catch( const std::exception& ex )
    {
        emit virtualMachinesFetchFailed( ex.what() );
    }

    _virtualMachinesFetchReply = nullptr;
    reply->deleteLater();
}

std::vector<std::shared_ptr<VirtualMachine>>
Node::processVirtualMachinesFetchReply( QNetworkReply* reply )
{
    try
    {
        const auto virtualMachineIDs = ProxmoxJsonParser::parseVirtualMachinesResponse( reply );

        std::vector<std::shared_ptr<VirtualMachine>> result;
        result.reserve( virtualMachineIDs.size() );

        for( const auto& vmid : virtualMachineIDs )
        {
            result.emplace_back( std::make_shared<VirtualMachine>( _restClient,
                                                                   vmid,
                                                                   _id ) );
        }

        return result;
    }
    catch( const VirtualMachineParsingException& ex )
    {
        throw NodeException( ex.what() );
    }
}

} // namespace Basealt
