#include "proxmox-rest-client.h"

#include <QNetworkCookie>

namespace {

const QString kProxmoxAPIInputDataMIMEType =
        "application/x-www-form-urlencoded";

const QString kAuthCookie = "PVEAuthCookie";
const QString kAuthHeader = "CSRFPreventionToken";

const QString kJsonAPIPath = "/api2/json";

QNetworkRequest makeRequest( const QUrl& url )
{
    QNetworkRequest request( url );

    request.setHeader( QNetworkRequest::ContentTypeHeader,
                       kProxmoxAPIInputDataMIMEType );

    request.setSslConfiguration( QSslConfiguration::defaultConfiguration() );

    return request;
}

void addAuthCookie( QNetworkRequest& request,
                    const std::shared_ptr<Basealt::User>& user )
{
    const auto cookie = QNetworkCookie( kAuthCookie.toUtf8(),
                                        user->ticket().toUtf8() );

    request.setHeader( QNetworkRequest::CookieHeader,
                       QVariant::fromValue( cookie ) );
}

} // namespace

namespace Basealt {

ProxmoxRestClient::ProxmoxRestClient( const QString& baseURL,
                                      bool ignoreSslErrors )
    : _baseURL( baseURL ),
      _user(),
      _networkManager(),
      _ignoreSslErrors( ignoreSslErrors )
{
    connect( &_networkManager,
             &QNetworkAccessManager::finished,
             this,
             &ProxmoxRestClient::finished );

    connect( &_networkManager,
             &QNetworkAccessManager::sslErrors,
             this,
             &ProxmoxRestClient::onSslErrors );

    _networkManager.setCookieJar( nullptr );
}

QNetworkReply* ProxmoxRestClient::get( const QString& endpoint )
{
    auto url = _baseURL;
    url.setPath( kJsonAPIPath + endpoint );

    auto request = makeRequest( url );

    if( const auto user = _user )
    {
        addAuthCookie( request, user );
    }

    return _networkManager.get( request );
}

QNetworkReply* ProxmoxRestClient::post( const QString& endpoint,
                                        const QByteArray& data )
{
    auto url = _baseURL;
    url.setPath( kJsonAPIPath + endpoint );

    auto request = makeRequest( url );
    request.setHeader( QNetworkRequest::ContentLengthHeader,
                       QByteArray::number( data.size() ) );

    if( const auto user = _user )
    {
        addAuthCookie( request, user );

        request.setRawHeader( kAuthHeader.toUtf8(), user->token().toUtf8() );
    }

    return _networkManager.post( request, data );
}

void ProxmoxRestClient::onSslErrors( QNetworkReply* reply,
                                     const QList<QSslError>& errors )
{
    if( _ignoreSslErrors )
    {
        reply->ignoreSslErrors();
    }
}

} // namespace Basealt
