#include "proxmox-virtual-machine.h"

#include <unordered_map>

#include <QUrlQuery>

#include "enum-class-hash.h"
#include "proxmox-json-parser.h"

namespace {

const QString kStateEndpointFormat = "/nodes/%1/qemu/%2/status/current";
const QString kChangeStatusEndpointFormat = "/nodes/%1/qemu/%2/status/%3";
const QString kSpiceProxyEndpointFormat = "/nodes/%1/qemu/%2/spiceproxy";
const QString kProxy = "proxy";

const std::unordered_map<Basealt::VirtualMachine::Status,
                         QString,
                         Basealt::EnumClassHash> kChangeStatusCommands
{
    { Basealt::VirtualMachine::Status::Running, "start" },
    { Basealt::VirtualMachine::Status::Stopped, "stop" }
};

const QString kInvalidSetStatusRespose = "Could not set status: invalid response";

} // namespace

namespace Basealt {

VirtualMachine::VirtualMachine( const std::shared_ptr<ProxmoxRestClient>& restClient,
                                size_t vmID,
                                const QString& nodeID )
    : _restClient( restClient ),
      _vmID( vmID ),
      _nodeID( nodeID ),
      _fetchStatusReply( nullptr ),
      _fetchSpiceProxyReply( nullptr ),
      _setStatusReply( nullptr )
{
    connect( _restClient.get(),
             &ProxmoxRestClient::finished,
             this,
             &VirtualMachine::onReplyFinished );
}

void VirtualMachine::fetchStatus()
{
    if( _fetchStatusReply )
    {
        return;
    }

    const auto url = QString( kStateEndpointFormat )
                     .arg( _nodeID, QString::number( _vmID ) );

    _fetchStatusReply = _restClient->get( url );
}

void VirtualMachine::fetchSpiceProxy()
{
    if( _fetchSpiceProxyReply )
    {
        return;
    }

    const auto url = QString( kSpiceProxyEndpointFormat )
                     .arg( _nodeID, QString::number( _vmID ) );

    const auto host = _restClient->url().host();

    QUrlQuery postData;
    postData.addQueryItem( kProxy, QUrl::toPercentEncoding( host ) );

    const auto data = postData.query().toUtf8();

    _fetchSpiceProxyReply = _restClient->post( url, data );
}

void VirtualMachine::setStatus( Status newState )
{
    if( _setStatusReply )
    {
        return;
    }

    const auto url = QString( kChangeStatusEndpointFormat )
                     .arg( _nodeID,
                           QString::number( _vmID ),
                           kChangeStatusCommands.at( newState ) );

    _setStatusReply = _restClient->post( url, QByteArray() );
}

void VirtualMachine::onReplyFinished( QNetworkReply* reply )
{
    if( _fetchStatusReply && _fetchStatusReply == reply )
    {
        try
        {
            emit statusFetched( ProxmoxJsonParser::parseVirtualMachineCurrentStatusResponse( reply ) );
        }
        catch( const std::exception& ex )
        {
            emit statusFetchFailed( ex.what() );
        }

        _fetchStatusReply = nullptr;
        reply->deleteLater();
        return;
    }

    if( _fetchSpiceProxyReply && _fetchSpiceProxyReply == reply )
    {
        try
        {
            emit spiceProxyFetched( ProxmoxJsonParser::parseGetSpiceProxyResponse( reply ) );
        }
        catch( const std::exception& ex )
        {
            emit spiceProxyFetchFailed( ex.what() );
        }

        _fetchSpiceProxyReply = nullptr;
        reply->deleteLater();
        return;
    }

    if( _setStatusReply && _setStatusReply == reply )
    {
        try
        {
            const auto response = ProxmoxJsonParser::parseVirtualMachineSetStatusResponse( reply );

            if( response.isEmpty() )
            {
                emit setStatusFailed( kInvalidSetStatusRespose );
            }
            else
            {
                emit statusSet();
            }
        }
        catch( const std::exception& ex )
        {
            emit setStatusFailed( ex.what() );
        }

        _setStatusReply = nullptr;
        reply->deleteLater();
        return;
    }
}

} // namespace Basealt
