#include "settings-view.h"

#include <QVBoxLayout>

namespace {

const QString kServerURLPlaceholder = "Server URL";
const QString kServerURLToolTip = "Server URL or IP address";

const QString kVirtualMachineIDPlaceholder = "Virtual machine ID";
const QString kVirtualMachineIDToolTip = "Virtual machine identifier you want to connect with";

const QString kUsernamePlaceholder = "Username";
const QString kUsernameToolTip = "Your username at the server";

const QString kPasswordPlaceholder = "Password";
const QString kPasswordToolTip = "Your password at the server";

const QString kRememberMeLabel = "Remember me";
const QString kRememberMeToolTip = "Restore settings on startup";

const QString kIgnoreSslErrorsLabel = "Ignore SSL errors";
const QString kIgnoreSslErrorsToolTip = "Do not stop connection on SSL error occured";

const QString kConnectButtonLabel = "Connect";
const QString kConnectButtonToolTip = "Connect to the virtual machine";

} // namespace

namespace Basealt {

SettingsView::SettingsView( QWidget* parent, Qt::WindowFlags flags )
    : QWidget( parent, flags ),
      _serverURLLineEdit( makeServerURLLineEdit() ),
      _virtualMachineIDLineEdit( makeVirtualMachineIDLineEdit() ),
      _usernameLineEdit( makeUsernameLineEdit() ),
      _passwordLineEdit( makePasswordLineEdit() ),
      _rememberMeCheckBox( makeRememberMeCheckBox() ),
      _ignoreSslErrorsCheckBox( makeIgnoreSslErrorsCheckBox() )
{
    const auto settingsWidgetLayout = new QVBoxLayout();
    settingsWidgetLayout->addWidget( _serverURLLineEdit );
    settingsWidgetLayout->addWidget( _virtualMachineIDLineEdit );
    settingsWidgetLayout->addWidget( _usernameLineEdit );
    settingsWidgetLayout->addWidget( _passwordLineEdit );
    settingsWidgetLayout->addWidget( _rememberMeCheckBox );
    settingsWidgetLayout->addWidget( _ignoreSslErrorsCheckBox );
    settingsWidgetLayout->addWidget( makeConnectPushButton(),
                                     0,
                                     Qt::AlignHCenter );

    setLayout( settingsWidgetLayout );
}

QLineEdit* SettingsView::makeServerURLLineEdit()
{
    auto serverURLLineEdit = new QLineEdit();

    serverURLLineEdit->setPlaceholderText( kServerURLPlaceholder );
    serverURLLineEdit->setToolTip( kServerURLToolTip );

    connect( serverURLLineEdit,
             &QLineEdit::editingFinished,
             this,
             &SettingsView::onServerURLEditingFinished );

    return serverURLLineEdit;
}

QLineEdit* SettingsView::makeVirtualMachineIDLineEdit()
{
    auto virtualMachineIDLineEdit = new QLineEdit();

    virtualMachineIDLineEdit->setPlaceholderText( kVirtualMachineIDPlaceholder );
    virtualMachineIDLineEdit->setToolTip( kVirtualMachineIDToolTip );

    connect( virtualMachineIDLineEdit,
             &QLineEdit::editingFinished,
             this,
             &SettingsView::onVirtualMachineIDEditingFinished );

    return virtualMachineIDLineEdit;
}

QLineEdit* SettingsView::makeUsernameLineEdit()
{
    auto usernameLineEdit = new QLineEdit();

    usernameLineEdit->setPlaceholderText( kUsernamePlaceholder );
    usernameLineEdit->setToolTip( kUsernameToolTip );

    connect( usernameLineEdit,
             &QLineEdit::editingFinished,
             this,
             &SettingsView::onUsernameEditingFinished );

    return usernameLineEdit;
}

QLineEdit* SettingsView::makePasswordLineEdit()
{
    auto passwordLineEdit = new QLineEdit();

    passwordLineEdit->setPlaceholderText( kPasswordPlaceholder );
    passwordLineEdit->setToolTip( kPasswordToolTip );
    passwordLineEdit->setEchoMode( QLineEdit::Password );

    connect( passwordLineEdit,
             &QLineEdit::editingFinished,
             this,
             &SettingsView::onPasswordEditingFinished );

    return passwordLineEdit;
}

QCheckBox* SettingsView::makeRememberMeCheckBox()
{
    auto rememberMeCheckBox = new QCheckBox( kRememberMeLabel );

    rememberMeCheckBox->setToolTip( kRememberMeToolTip );

    connect( rememberMeCheckBox,
             &QCheckBox::stateChanged,
             this,
             &SettingsView::onRememberMeToggled );

    return rememberMeCheckBox;
}

QCheckBox* SettingsView::makeIgnoreSslErrorsCheckBox()
{
    auto ignoreSslErrorsCheckBox = new QCheckBox( kIgnoreSslErrorsLabel );

    ignoreSslErrorsCheckBox->setToolTip( kIgnoreSslErrorsToolTip );

    connect( ignoreSslErrorsCheckBox,
             &QCheckBox::stateChanged,
             this,
             &SettingsView::onIgnoreSslErrorsToggled );

    return ignoreSslErrorsCheckBox;
}

QPushButton* SettingsView::makeConnectPushButton()
{
    auto connectPushButton = new QPushButton( kConnectButtonLabel );

    connectPushButton->setToolTip( kConnectButtonToolTip );

    connect( connectPushButton,
             &QPushButton::clicked,
             this,
             &SettingsView::connectButtonClicked );

    return connectPushButton;
}

void SettingsView::onServerURLEditingFinished()
{
    emit serverURLChanged( _serverURLLineEdit->text() );
}

void SettingsView::onVirtualMachineIDEditingFinished()
{
    emit virtualMachineIDChanged( _virtualMachineIDLineEdit->text() );
}

void SettingsView::onUsernameEditingFinished()
{
    emit usernameChanged( _usernameLineEdit->text() );
}

void SettingsView::onPasswordEditingFinished()
{
    emit passwordChanged( _passwordLineEdit->text() );
}

void SettingsView::onRememberMeToggled( int state )
{
    emit rememberMeStateChanged( state == Qt::Checked );
}

void SettingsView::onIgnoreSslErrorsToggled( int state )
{
    emit ignoreSslErrorsStateChanged( state == Qt::Checked );
}

} // namespace Basealt
